//
//	Bitmap
//

#ifndef		_BITMAP_INC_
#define		_BITMAP_INC_

#include "image.hpp"
#include "stream.hpp"

#include <string.h>
#include <assert.h>

class BitmapException : public ImageException
{
protected:
	BitmapException(const char *name, const char *desc, int line = 0)
		: ImageException(name, desc, line)
	{
	}
public:
	BitmapException(const char *desc = "unknown", int line = 0)
		: ImageException("BitmapException", desc, line)
	{
	}
	virtual ~BitmapException()
	{
	}
};

class Bitmap : public Image
{
	Bitmap(Bitmap const &);
	Bitmap & operator = (Bitmap const &);
public:
	enum Mode { READ = 1, WRITE = 2 };
protected:
	#pragma pack(push, 1)
	struct BMP_HEAD
	{
		word magic;
		uint filesize;
		uint reserved;
		uint offset;
	};
	struct BMP_INFO
	{
		uint size_struct;
		long width;
		long height;
		word planes;
		word bit_count;
		uint compression;
		uint size_image;
		long x_pels_per_meter;
		long y_pels_per_meter;
		uint color_used;
		uint color_important;
	};
	#pragma pack(pop)
	enum
	{
		BMI_BI_RGB       = 0,
		BMI_BI_BITFIELDS = 3,
	};
protected:
	static uint stride_(uint w, uint d)
	{
		return ((w * d + 31) / 8) & ~3;
	}
	void init_()
	{
		strm_ = 0;
		mask_ = 0;
		plte_ = 0;
		bits_ = 0;
		scan_ = 0;
		memset(&info_, 0, sizeof info_);
	}
public:
	Bitmap()
	{
		init_();
	}
	explicit Bitmap(Stream &s, Mode m = READ)
	{	// flag (READ | WRITE) is not available...
		assert((m & (READ | WRITE)) != (READ | WRITE));
		init_();
		if (m & READ) {
			BMP_HEAD head;
			s.read(head);
			if (head.magic != *(const word*)"BM")
				throw BitmapException("invalid format");
			try {
				s.read(info_);
				if (info_.size_struct != sizeof info_ ||
				    info_.width  <= 0 ||
				    info_.height == 0 ||
				    info_.planes != 1) {
					throw BitmapException("invalid format");
				}
				switch (info_.bit_count) {
				case 1: case 4: case 8: case 24:
					if (info_.compression == BMI_BI_RGB)
						break;
					throw BitmapException("invalid format");
				case 16: case 32:
					if (info_.compression == BMI_BI_RGB)
						break;
					if (info_.compression == BMI_BI_BITFIELDS) {
						mask_ = new uint[3];
						s.readExact(mask_, sizeof*mask_ * 3);
						break;
					}
					throw BitmapException("invalid format");
				default:
					throw BitmapException("invalid format");
				}
				if (info_.bit_count <= 8) {
					uint len = 1UL << info_.bit_count;
					plte_ = new uint[len];
					memset(plte_, 0, sizeof*plte_ * len);
					s.readExact(plte_, sizeof*plte_ * (info_.color_used ? info_.color_used : len));
				} else if (info_.color_used) {
					plte_ = new uint[info_.color_used];
					s.readExact(plte_, sizeof*plte_ * info_.color_used);
				}
				bool topdown = info_.height < 0;
				if (topdown)
					info_.height = -info_.height;
				uint n = stride_(info_.width, info_.bit_count);
				bits_ = new byte[n * info_.height];
				scan_ = new Scanline[info_.height];
				for (long y = 0; y < info_.height; y++)
					scan_[y](bits_ + n * (info_.height - 1 - y), info_.width, info_.bit_count, n, mask_);
				if (topdown) {
					for (long y = 0; y < info_.height; y++)
						s.readExact(scan_[y].ptr, scan_[y].length);
				} else {
					for (long y = info_.height - 1; y >= 0; y--)
						s.readExact(scan_[y].ptr, scan_[y].length);
				}
			} catch (Exception e) {
				destroy();
				throw e;
			} catch (...) {
				destroy();
				throw BitmapException("out of memory");
			}
		}
		if (m & WRITE)
			strm_ = &s;
	}
	Bitmap(uint w, uint h, uint d = 24, const uint *p = 0, uint c = 0, const uint *m = 0)
	{
		init_();
		create(w, h, d, p, c, m);
	}
	Scanline operator [] (uint y)
	{
		return scan_[y];
	}
	void create(uint w, uint h, uint d = 24, const uint *p = 0, uint c = 0, const uint *m = 0)
	{
		destroy();
		if (w == 0 || h == 0)
			return;
		info_.width      = w;
		info_.height     = h;
		info_.bit_count  = (word)d;
		info_.color_used = 0;
		try {
			if (d <= 8) {
				uint len = 1UL << d;
				assert(c <= len);
				plte_ = new uint[len];
				memset(plte_, 0, sizeof(uint) * len);
				memcpy(plte_, p, sizeof(uint) * (c ? c : len));
				info_.color_used = c;
			} else if (p && c != 0) {
				plte_ = new uint[c];
				memcpy(plte_, p, sizeof(uint) * c);
				info_.color_used = c;
			}
			if (d == 16 || d == 32) {
				if (m) {
					mask_ = new uint[3];
					memcpy(mask_, m, sizeof(uint) * 3);
				}
			}
			uint n = stride_(w, d);
			bits_ = new byte[n * h];
			scan_ = new Scanline[h];
			for (uint y = 0; y < h; y++)
				scan_[y](bits_ + n * (h - 1 - y), w, d, n, m);
		} catch (...) {
			destroy();
			throw BitmapException("out of memory");
		}
	}
	void flush()
	{
		if (strm_)
			flush(*strm_);
	}
	void flush(Stream &s)
	{
		if (!scan_)
			return;
		info_.size_image  = stride_(info_.width, info_.bit_count) * info_.height;
		switch (info_.bit_count) {
		case 1: case 4: case 8: case 24:
			info_.compression = BMI_BI_RGB;
			break;
		case 16: case 32:
			info_.compression = mask_ ? BMI_BI_BITFIELDS : BMI_BI_RGB;
			break;
		}
		uint count_palette = info_.bit_count <= 8
			? info_.color_used == 0
				? 1UL << info_.bit_count
				: info_.color_used
			: info_.color_used;
		BMP_HEAD head = { *(const word*)"BM" };
		head.offset = sizeof(head) + sizeof(info_);
		if (info_.compression == BMI_BI_BITFIELDS)
			head.offset += sizeof(uint) * 3;
		head.offset += sizeof(uint) * count_palette;
		head.filesize = head.offset + info_.size_image;
		s.write(head);
		info_.size_struct      = sizeof(info_);
		info_.planes           = 1;
		info_.y_pels_per_meter = 0;
		info_.x_pels_per_meter = 0;
		info_.color_important  = 0;
		s.write(info_);
		if (info_.compression == BMI_BI_BITFIELDS)
			s.writeExact(mask_, sizeof*mask_ * 3);
		if (count_palette != 0) {
			assert(plte_);
			s.writeExact(plte_, sizeof*plte_ * count_palette);
		}
		for (long y = info_.height - 1; y >= 0; y--)
			s.writeExact(scan_[y].ptr, scan_[y].length);
	}
	void destroy()
	{
		delete[] mask_; mask_ = 0;
		delete[] plte_; plte_ = 0;
		delete[] bits_; bits_ = 0;
		delete[] scan_; scan_ = 0;
		memset(&info_, 0, sizeof info_);
	}
	virtual ~Bitmap()
	{
		flush();
		destroy();
	}
protected:
	Stream *    strm_;
	BMP_INFO    info_;
	uint *      mask_;
	uint *      plte_;
	byte *      bits_;
	Scanline *  scan_;
	
	uint width_get() const
	{
		return info_.width;
	}
	uint height_get() const
	{
		return info_.height;
	}
	uint depth_get() const
	{
		return info_.bit_count;
	}
	uint &palette_arr(uint const &i)
	{
		return plte_[i];
	}
	uint *palette_get() const
	{
		return plte_;
	}
};

#endif	//	_BITMAP_INC_
