//
//	bmp2png
//

#include "bitmap.hpp"
#include "pngdef.hpp"
#include "shared_ptr.hpp"

#include <stdio.h>

typedef unsigned char byte;
typedef unsigned long uint;

uint swap_endian(uint x)
{
	return x << 24 | x << 8 & 0xFF0000 | x >> 8 & 0xFF00 | x >> 24;
}

void put_chunk(Stream &png, const char *name, Stream &data, uint len)
{
	static CRC32 crc;
	crc.init().puts(name, 4);
	png.put(swap_endian(len));
	png.writeExact(name, 4);
	for (uint i = 0; i < len; i++) {
		byte x = data.get<byte>();
		crc.put(x);
		png.put(x);
	}
	png.put(swap_endian(crc));
}
void put_chunk(Stream &png, const char *name, const void *data, uint len)
{
	static CRC32 crc;
	crc.init().puts(name, 4).puts(data, len);
	png.put(swap_endian(len));
	png.writeExact(name, 4);
	png.writeExact(data, len);
	png.put(swap_endian(crc));
}
template <class T>
void put_chunk(Stream &png, const char *name, T const &data)
{
	put_chunk(png, name, &data, sizeof data);
}

int main(int argc, char *argv[])
{
	for (int i = 1; i < argc; i++) {
		try {
			const char *arg = argv[i];
			const char *ext = strrchr(arg, '.');
			File ifs(arg);
			Bitmap bmp(ifs);
			char out[FILENAME_MAX];
			if (!ext || strcmp(ext, ".png") == 0)
				sprintf(out, "%s.png", arg);
			else
				sprintf(out, "%.*s.png", ext - arg, arg);
			File png(out, File::WRITE);
			printf("%s -> %s\n", arg, out);
			png.write(PNG::SIGNATURE);
			PNG::IHDR ihdr = {
				swap_endian(bmp.width),
				swap_endian(bmp.height),
				8,
			};
			if (bmp.depth <= 8) {
				ihdr.color_type = PNG::IHDR::COLOR_PLTE;
				put_chunk(png, "IHDR", ihdr);
				uint npal = 1U << bmp.depth;
				PNG::PLTE pal[256] = {0};
				for (uint i = 0; i < npal; i++) {
					uint bgr = bmp.palette[i];
					pal[i].r = bgr >> 16 & 0xFF;
					pal[i].g = bgr >>  8 & 0xFF;
					pal[i].b = bgr       & 0xFF;
				}
				put_chunk(png, "PLTE", pal);
				MemoryStream mems;
				PNG::Z_HDR zhdr(PNG::Z_HDR::NO_COMPRESSION);
				mems.put(zhdr);
				Adler32 adler32;
				for (uint y = 0; y < bmp.height; y++) {
					PNG::Z_BHDR bhdr(
						y == bmp.height - 1,
						1 + bmp.width
					);
					mems.put(bhdr);
					byte filter = 0;
					adler32.put(filter), mems.put(filter);
					for (uint x = 0; x < bmp.width; x++) {
						byte t = (byte)bmp[y][x];
						adler32.put(t), mems.put(t);
					}
				}
				mems.put(swap_endian(adler32));
				uint len = mems.position;
				mems.rewind();
				put_chunk(png, "IDAT", mems, len);
			} else if (bmp.depth != 32) {
				ihdr.color_type = PNG::IHDR::COLOR_RGB;
				put_chunk(png, "IHDR", ihdr);
				MemoryStream mems;
				PNG::Z_HDR zhdr(PNG::Z_HDR::NO_COMPRESSION);
				mems.put(zhdr);
				Adler32 adler32;
				for (uint y = 0; y < bmp.height; y++) {
					PNG::Z_BHDR bhdr(
						y == bmp.height - 1,
						1 + bmp.width * 3
					);
					mems.put(bhdr);
					byte filter = 0;
					adler32.put(filter), mems.put(filter);
					for (uint x = 0; x < bmp.width; x++) {
						uint t = bmp[y][x];
						byte r = byte(t >> 16);
						byte g = byte(t >>  8);
						byte b = byte(t      );
						adler32.put(r), mems.put(r);
						adler32.put(g), mems.put(g);
						adler32.put(b), mems.put(b);
					}
				}
				mems.put(swap_endian(adler32));
				uint len = mems.position;
				mems.rewind();
				put_chunk(png, "IDAT", mems, len);
			} else {
				ihdr.color_type = PNG::IHDR::COLOR_RGBA;
				put_chunk(png, "IHDR", ihdr);
				MemoryStream mems;
				PNG::Z_HDR zhdr(PNG::Z_HDR::NO_COMPRESSION);
				mems.put(zhdr);
				Adler32 adler32;
				for (uint y = 0; y < bmp.height; y++) {
					PNG::Z_BHDR bhdr(
						y == bmp.height - 1,
						1 + bmp.width * 4
					);
					mems.put(bhdr);
					byte filter = 0x00;
					adler32.put(filter), mems.put(filter);
					uint *line = (uint *)(void *)bmp[y].ptr;
					for (uint x = 0; x < bmp.width; x++) {
						uint t = line[x];
						byte r = byte(t >> 16);
						byte g = byte(t >>  8);
						byte b = byte(t      );
						byte a = byte(t >> 24);
						adler32.put(r), mems.put(r);
						adler32.put(g), mems.put(g);
						adler32.put(b), mems.put(b);
						adler32.put(a), mems.put(a);
					}
				}
				mems.put(swap_endian(adler32));
				uint len = mems.position;
				mems.rewind();
				put_chunk(png, "IDAT", mems, len);
			}
			put_chunk(png, "IEND", NULL, 0);
		} catch (Exception e) {
			fprintf(stderr, "# %s: %s\n", e.name(), e.desc());
		}
	}
	return 0;
}
