//
//	Image
//

#ifndef		_IMAGE_INC_
#define		_IMAGE_INC_

#include "exception.hpp"
#include "property.hpp"

#include <string.h>
#include <assert.h>

class ImageException : public Exception
{
protected:
	ImageException(const char *name, const char *desc, int line = 0)
		: Exception(name, desc, line)
	{
	}
public:
	ImageException(const char *desc = "unknown", int line = 0)
		: Exception("ImageException", desc, line)
	{
	}
	virtual ~ImageException()
	{
	}
};

class Image
{
protected:
	typedef unsigned char  byte;
	typedef unsigned short word;
	typedef unsigned long  uint;
public:
	static const uint MASK_16_555[3];
	static const uint MASK_16_565[3];
	static const uint MASK_32_888[3];
	static const uint *MASK_16_DEFAULT;
	static const uint *MASK_32_DEFAULT;
public:
	class Scanline
	{
	public:
		property<void *, Scanline> ptr;
		property<uint, Scanline> length;
		property<uint, Scanline> width;
		property<uint, Scanline> depth;
	protected:
		void init_()
		{
			ptr(this, &Scanline::ptr_get);
			length(this, &Scanline::length_get);
			width(this, &Scanline::width_get);
			depth(this, &Scanline::depth_get);
		}
	public:
		Scanline()
		{
			init_();
			ptr_    = 0;
			length_ = 0;
			width_  = 0;
			depth_  = 0;
			mask_   = 0;
		}
		Scanline(Scanline const &s)
		{
			init_();
			ptr_    = s.ptr_;
			length_ = s.length_;
			width_  = s.width_;
			depth_  = s.depth_;
			mask_   = s.mask_;
		}
		Scanline(void *p, uint w, uint d, uint s, const uint *m = 0)
		{
			init_();
			operator () (p, w, d, s, m);
		}
		void operator () (void *p, uint w, uint d, uint s, const uint *m = 0)
		{
			ptr_    = (byte *)p;
			length_ = s;
			width_  = w;
			depth_  = d;
			switch (d) {
			case 1: case 4: case 8: case 24:
				mask_ = 0;
				break;
			case 16:
				mask_ = m ? m : MASK_16_DEFAULT;
				break;
			case 32:
				mask_ = m ? m : MASK_32_DEFAULT;
				break;
			default:
				assert(0);
			}
		}
		property_param<uint, Scanline, uint> operator [] (uint x)
		{
			return property_param<uint, Scanline, uint>(this, &Scanline::index_get, &Scanline::index_set, x);
		}
		virtual ~Scanline()
		{
		}
	protected:
		byte *ptr_;
		void *ptr_get() const
		{
			return ptr_;
		}
		uint length_;
		uint length_get() const
		{
			return length_;
		}
		uint width_;
		uint width_get() const
		{
			return width_;
		}
		uint depth_;
		uint depth_get() const
		{
			return depth_;
		}
		uint index_get(uint const &x) const
		{
			switch (depth_) {
			case  1: return ptr_[x / 8] >>  (~x & 7)      &  1;
			case  4: return ptr_[x / 2] >> ((~x & 1) * 4) & 15;
			case  8: return ptr_[x];
			case 24: return RGB_(ptr_[x * 3 + 2], ptr_[x * 3 + 1], ptr_[x * 3]);
			case 16: return UNMASK_(*(word *)(ptr_ + x * 2));
			case 32: return UNMASK_(*(uint *)(ptr_ + x * 4));
			default: break;
			}
			assert(0);
			return 0;
		}
		uint index_set(uint const &v, uint const &x)
		{
			switch (depth_) {
			case  1:
				ptr_[x / 8] = byte(
					ptr_[x / 8] & ~( 1 <<  (~x & 7)     ) | (v &  1) <<  (~x & 7)
				);
				return v;
			case  4:
				ptr_[x / 2] = byte(
					ptr_[x / 2] & ~(15 << ((~x & 1) * 4)) | (v & 15) << ((~x & 1) * 4)
				);
				return v;
			case  8:
				return ptr_[x] = (byte)v;
			case 24:
				memcpy(ptr_ + x * 3, &v, 3);
				return v;
			case 16:
				*(word *)(ptr_ + x * 2) = (word)MASK_(v);
				return v;
			case 32:
				*(uint *)(ptr_ + x * 4) = MASK_(v);
				return v;
			default:
				break;
			}
			assert(0);
			return 0;
		}
		const uint *mask_;
	protected:
		inline static uint RGB_(uint r, uint g, uint b) { return r << 16 | g << 8 | b; }
		inline static byte GET_R_(uint rgb) { return byte(rgb >> 16); }
		inline static byte GET_G_(uint rgb) { return byte(rgb >>  8); }
		inline static byte GET_B_(uint rgb) { return byte(rgb      ); }
		inline uint UNMASK_(uint x) const
		{
			uint r = x & mask_[0], g = x & mask_[1], b = x & mask_[2];
			return RGB_(
				r >> bsf(mask_[0]) << (8 - (bsr(mask_[0]) - bsf(mask_[0]) + 1)),
				g >> bsf(mask_[1]) << (8 - (bsr(mask_[1]) - bsf(mask_[1]) + 1)),
				b >> bsf(mask_[2]) << (8 - (bsr(mask_[2]) - bsf(mask_[2]) + 1))
			);
		}
		inline uint MASK_(uint x) const
		{
			uint r = GET_R_(x), g = GET_G_(x), b = GET_B_(x);
			return (
				r >> (8 - (bsr(mask_[0]) - bsf(mask_[0]) + 1)) << bsf(mask_[0]) |
				g >> (8 - (bsr(mask_[1]) - bsf(mask_[1]) + 1)) << bsf(mask_[1]) |
				b >> (8 - (bsr(mask_[2]) - bsf(mask_[2]) + 1)) << bsf(mask_[2])
			);
		}
		inline static uint bsf(uint x)
		{
			#if defined(NDEBUG) && defined(_MSC_VER)
				__asm bsf eax, x;
			#elif defined(NDEBUG) && defined(__DMC__)
				asm bsf eax, x;
			#elif defined(NDEBUG) && defined(__GNUC__)
				asm ("bsf %1, %0" : "=q"(x) : "q"(x));
				return x;
			#else
				for (int i = 0; i < 32; i++) {
					if (1UL << i & x)
						return i;
				}
				return 0;
			#endif
		}
		inline static uint bsr(uint x)
		{
			#if defined(NDEBUG) && defined(_MSC_VER)
				__asm bsr eax, x;
			#elif defined(NDEBUG) && defined(__DMC__)
				asm bsr eax, x;
			#elif defined(NDEBUG) && defined(__GNUC__)
				asm ("bsr %1, %0" : "=q"(x) : "q"(x));
				return x;
			#else
				for (int i = 31; i >= 0; i--) {
					if (1UL << i & x)
						return i;
				}
				return 0;
			#endif
		}
	};
public:
	property<uint, Image> width;
	property<uint, Image> height;
	property<uint, Image> depth;
	property_array<uint, Image, uint> palette;
public:
	Image()
	{
		width(this, &Image::width_get);
		height(this, &Image::height_get);
		depth(this, &Image::depth_get);
		palette(this, &Image::palette_arr, &Image::palette_get);
	}
	virtual Scanline operator [] (uint y) = 0;
	virtual ~Image()
	{
	}
protected:
	virtual uint width_get() const = 0;
	virtual uint height_get() const = 0;
	virtual uint depth_get() const = 0;
	virtual uint &palette_arr(uint const &i) = 0;
	virtual uint *palette_get() const = 0;
};

const Image::uint Image::MASK_16_555[3] = { 0x00007C00, 0x000003E0, 0x0000001F };
const Image::uint Image::MASK_16_565[3] = { 0x0000F800, 0x000007E0, 0x0000001F };
const Image::uint Image::MASK_32_888[3] = { 0x00FF0000, 0x0000FF00, 0x000000FF };
const Image::uint *Image::MASK_16_DEFAULT = Image::MASK_16_555;
const Image::uint *Image::MASK_32_DEFAULT = Image::MASK_32_888;

#endif	//	_IMAGE_INC_
