//
//	shared_ptr
//

#ifndef		SHARED_PTR_HPP
#define		SHARED_PTR_HPP

#include <stdlib.h>

struct shared_ptr_destroyer
{
	template <typename T>
	static void destroy(T *ptr)
	{
		delete ptr;
	}
};
struct shared_ary_destroyer
{
	template <typename T>
	static void destroy(T *ary)
	{
		delete[] ary;
	}
};
struct shared_mem_destroyer
{
	template <typename T>
	static void destroy(T *mem)
	{
		::free(mem);
	}
};

template <typename T>
struct shared_ptr_reference
{
	typedef T & t;
};
template <>
struct shared_ptr_reference<void>
{
	typedef void t;
};

template <typename T, class D = shared_ptr_destroyer>
class shared_ptr
{
	#if defined(__GNUC__) || defined(_MSC_VER)
		#define SHARED_PTR_INLINE inline
	#else
		#define SHARED_PTR_INLINE
	#endif
public:
	typedef typename shared_ptr_reference<T>::t reference_t;
	typedef shared_ptr<T, D> this_t;
	typedef T impl_t;
private:
	typedef T * this_t::*bool_t;
	this_t * prev_;
	this_t * next_;
	T * ptr_;
public:
	shared_ptr(T *p = 0)
		: ptr_(0)
	{	prev_ = next_ = this;
		operator = (p);
	}
	shared_ptr(this_t const &s)
		: ptr_(0)
	{	prev_ = next_ = this;
		operator = (s);
	}
	this_t & operator = (this_t const &s)
	{
		if (&s == this)
			return *this;
		_destroy();
		this_t *p = const_cast<this_t *>(&s);
		ptr_ = s.ptr_;
		prev_ = p;
		next_ = p->next_;
		p->next_ = next_->prev_ = this;
		return *this;
	}
	this_t & operator = (T *p)
	{
		if (p == ptr_)
			return *this;
		_destroy();
		ptr_ = p;
		return *this;
	}
	reference_t operator [] (int i) const
	{
		return ptr_[i];
	}
	reference_t operator * () const
	{
		return *ptr_;
	}
	T * operator -> () const
	{
		return ptr_;
	}
	operator bool_t () const
	{
		return !ptr_ ? 0 : &this_t::ptr_;
	}
	bool operator ! () const
	{
		return !ptr_;
	}
	inline friend bool operator == (this_t const &lhs, const T *rhs)
	{
		return lhs.ptr_ == rhs;
	}
	inline friend bool operator != (this_t const &lhs, const T *rhs)
	{
		return lhs.ptr_ != rhs;
	}
	inline friend bool operator == (const T *lhs, this_t const &rhs)
	{
		return lhs == rhs.ptr_;
	}
	inline friend bool operator != (const T *lhs, this_t const &rhs)
	{
		return lhs != rhs.ptr_;
	}
	template <typename U>
	SHARED_PTR_INLINE friend bool operator == (this_t const &lhs, const U *rhs)
	{
		return lhs.ptr_ == rhs;
	}
	template <typename U>
	SHARED_PTR_INLINE friend bool operator != (this_t const &lhs, const U *rhs)
	{
		return lhs.ptr_ != rhs;
	}
	template <typename U>
	SHARED_PTR_INLINE friend bool operator == (const U *lhs, this_t const &rhs)
	{
		return lhs == rhs.ptr_;
	}
	template <typename U>
	SHARED_PTR_INLINE friend bool operator != (const U *lhs, this_t const &rhs)
	{
		return lhs != rhs.ptr_;
	}
	template <typename U, class E>
	bool operator == (shared_ptr<U, E> const &s) const
	{
		return ptr_ == s.ptr_;
	}
	template <typename U, class E>
	bool operator != (shared_ptr<U, E> const &s) const
	{
		return ptr_ != s.ptr_;
	}
	T * _getimpl() const
	{
		return ptr_;
	}
	bool _unique() const
	{
		return prev_ == this;
	}
	T * _release()
	{
		if (_unique()) {
			T *p = ptr_;
			ptr_ = 0;
			return p;
		}
		prev_->next_ = next_;
		next_->prev_ = prev_;
		prev_ = next_ = this;
		return ptr_ = 0;
	}
	T * _usurp()
	{
		T *p = ptr_;
		ptr_ = 0;
		for (this_t *i = next_; i != this; ) {
			this_t *next = i->next_;
			i->prev_ = i->next_ = i;
			i->ptr_ = 0;
			i = next;
		}
		return p;
	}
	void _destroy()
	{
		T *p = _release();
	#ifndef SHARED_PTR_ALLOW_DESTROY_NULL
		if (p)
	#endif
		D::destroy(p);
	}
	~shared_ptr()
	{
		_destroy();
	}
private:
	#undef SHARED_PTR_INLINE
};

#endif	//	SHARED_PTR_HPP
