//
//	Stream
//

#ifndef		_STREAM_INC_
#define		_STREAM_INC_

#include "exception.hpp"
#include "property.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

class StreamException : public Exception
{
protected:
	StreamException(const char *name, const char *desc, int line = 0)
		: Exception(name, desc, line)
	{
	}
public:
	StreamException(const char *desc = "unknown", int line = 0)
		: Exception("StreamException", desc, line)
	{
	}
	virtual ~StreamException()
	{
	}
};

class InputStream
{
public:
	virtual size_t readBlock(void *, size_t) = 0;
	void readExact(void *buffer, size_t length)
	{
		if (readBlock(buffer, length) != length)
			throw StreamException("readExact");
	}
	template <class T>
	void read(T &x)
	{
		readExact(&x, sizeof x);
	}
	template <typename T>
	T get()
	{
		T x;
		read(x);
		return x;
	}
	virtual ~InputStream()
	{
	}
};

class OutputStream
{
public:
	virtual size_t writeBlock(const void *, size_t) = 0;
	void writeExact(const void *buffer, size_t length)
	{
		if (writeBlock(buffer, length) != length)
			throw StreamException("writeExact");
	}
	template <class T>
	void write(T const &x)
	{
		writeExact(&x, sizeof x);
	}
	template <typename T>
	void put(T x)
	{
		write(x);
	}
	virtual ~OutputStream()
	{
	}
};

class Stream : public InputStream, public OutputStream
{
	Stream(Stream const &);
	Stream & operator = (Stream const &);
public:
	property<bool, Stream> readable;
	property<bool, Stream> writeable;
	property<bool, Stream> seekable;
	enum SeekOrigin
	{
		SET = 0,
		CUR = 1,
		END = 2,
	};
	property<unsigned long, Stream> position;
	property<unsigned long, Stream> length;
public:
	Stream()
		: readable_(false), writeable_(false), seekable_(false)
	{
		readable(this, &Stream::readable_get);
		writeable(this, &Stream::writeable_get);
		seekable(this, &Stream::seekable_get);
		position(this,
			reinterpret_cast<unsigned long (Stream::*)() const>(&Stream::position_get),
			&Stream::position_set);
		length(this,
			reinterpret_cast<unsigned long (Stream::*)() const>(&Stream::length_get));
	}
	virtual unsigned long seek(long, SeekOrigin) = 0;
	void rewind()
	{
		assert(seekable_);
		seek(0, SET);
	}
	size_t copyFrom(Stream &s)
	{
		size_t len = 0;
		char buf[8192];
		for (size_t n; (n = s.readBlock(buf, sizeof buf)) != 0; len += n)
			writeExact(buf, n);
		return len;
	}
	size_t copyFrom(Stream &s, size_t length)
	{
		size_t len = 0;
		char buf[8192];
		for (size_t n, t; length != 0 && (n = s.readBlock(buf, t
			= length < sizeof buf ? length : sizeof buf)) != 0; len += n, length -= n) {
			writeExact(buf, n);
			if (n != t)
				break;
		}
		return len;
	}
	virtual ~Stream()
	{
	}
protected:
	bool readable_;
	bool readable_get() const
	{
		return readable_;
	}
	bool writeable_;
	bool writeable_get() const
	{
		return writeable_;
	}
	bool seekable_;
	bool seekable_get() const
	{
		return seekable_;
	}
	unsigned long position_get()
	{
		assert(seekable_);
		return seek(0, CUR);
	}
	unsigned long position_set(unsigned long const &p)
	{
		assert(seekable_);
		return seek(p, SET);
	}
	unsigned long length_get()
	{
		assert(seekable_);
		unsigned long p = seek(0, CUR);
		unsigned long r = seek(0, END);
		seek(p, SET);
		return r;
	}
};

class FileException : public StreamException
{
protected:
	FileException(const char *name, const char *desc, int line = 0)
		: StreamException(name, desc, line)
	{
	}
public:
	FileException(const char *desc = "unknown", int line = 0)
		: StreamException("FileException", desc, line)
	{
	}
	virtual ~FileException()
	{
	}
};

class File : public Stream
{
	File(File const &);
	File & operator = (File const &);
public:
	enum Mode
	{
		READ  = 1,
		WRITE = 2
	};
public:
	File()
		: Stream(), fp_(NULL)
	{
	}
	explicit File(const char *name, Mode mode = READ)
		: Stream(), fp_(NULL)
	{
		assert(name != 0);
		open(name, mode);
	}
	void open(const char *name, Mode mode = READ)
	{
		assert(name != 0);
		close();
		switch (mode) {
		case READ:
			fp_ = fopen(name, "rb");
			break;
		case WRITE:
			fp_ = fopen(name, "wb");
			break;
		case READ | WRITE:
			fp_ = fopen(name, "r+b");
			break;
		default:
			assert(0);
		}
		if (fp_ == NULL)
			throw FileException("open");
		readable_  = (mode & READ ) != 0;
		writeable_ = (mode & WRITE) != 0;
		seekable_  = fseek(fp_, 0, SEEK_SET) == 0;
	}
	size_t readBlock(void *buffer, size_t length)
	{
		assert(readable_);
		return fread(buffer, 1, length, fp_);
	}
	size_t writeBlock(const void *buffer, size_t length)
	{
		assert(writeable_);
		return fwrite(buffer, 1, length, fp_);
	}
	unsigned long seek(long offset, SeekOrigin origin)
	{
		assert(seekable_);
		if (fseek(fp_, offset, origin) != 0)
			throw FileException("seek");
		long r = ftell(fp_);
		if (r == -1)
			throw FileException("seek");
		return static_cast<unsigned long>(r);
	}
	void flush()
	{
		if (fp_ != NULL)
			fflush(fp_);
	}
	void close()
	{
		if (fp_ != NULL) {
			fclose(fp_);
			fp_  = NULL;
		}
		readable_  = false;
		writeable_ = false;
		seekable_  = false;
	}
	virtual ~File()
	{
		close();
	}
protected:
	FILE *fp_;
};

class MemoryStreamException : public StreamException
{
protected:
	MemoryStreamException(const char *name, const char *desc, int line = 0)
		: StreamException(name, desc, line)
	{
	}
public:
	MemoryStreamException(const char *desc = "unknown", int line = 0)
		: StreamException("MemoryStreamException", desc, line)
	{
	}
	virtual ~MemoryStreamException()
	{
	}
};

class MemoryStream : public Stream
{
protected:
	typedef unsigned char byte;
public:
	MemoryStream()
		: Stream(), reference_(false), ptr_(0), cursor_(0), length_(0), capacity_(0)
	{
		readable_  = true;
		seekable_  = true;
		writeable_ = true;
	}
	MemoryStream(void *ptr, size_t length, bool allow_write = true)
		: Stream(), reference_(false), ptr_(0), cursor_(0), length_(0), capacity_(0)
	{
		readable_  = true;
		seekable_  = true;
		writeable_ = allow_write;
		reference_ = true;
		ptr_       = reinterpret_cast<unsigned char *>(ptr);
		length_    = length;
	}
	MemoryStream(const void *ptr, size_t length)
		: Stream(), reference_(false), ptr_(0), cursor_(0), length_(0), capacity_(0)
	{
		readable_  = true;
		seekable_  = true;
		writeable_ = false;
		reference_ = true;
		ptr_       = reinterpret_cast<unsigned char *>(const_cast<void *>(ptr));
		length_    = length;
	}
	MemoryStream(MemoryStream const &ms)
		: Stream(), reference_(false), ptr_(0), cursor_(0), length_(0), capacity_(0)
	{
		readable_  = true;
		seekable_  = true;
		try {
			*this = ms;
		} catch (MemoryStreamException e) {
			clear();
			throw e;
		}
	}
	MemoryStream & operator = (MemoryStream const &ms)
	{
		clear();
		if (ms.reference_) {
			reference_ = true;
			ptr_       = ms.ptr_;
			cursor_    = ms.cursor_;
			length_    = ms.length_;
			writeable_ = ms.writeable_;
		} else if (ms.length_ != 0) {
			expand_(ms.length_);
			memcpy(ptr_, ms.ptr_, ms.length_);
			cursor_    = ms.cursor_;
		}
		return *this;
	}
	void reserve(size_t n)
	{
		expand_(n);
	}
	size_t readBlock(void *buffer, size_t length)
	{
		if (length > length_ - cursor_)
			length = length_ - cursor_;
		memcpy(buffer, ptr_ + cursor_, length);
		cursor_ += length;
		return length;
	}
	size_t writeBlock(const void *buffer, size_t length)
	{
		assert(writeable_);
		if (length > length_ - cursor_) {
			if (reference_) {
				length = length_ - cursor_;
				memcpy(ptr_ + cursor_, buffer, length);
				cursor_ += length;
				return length;
			}
			expand_(cursor_ + length);
		}
		memcpy(ptr_ + cursor_, buffer, length);
		cursor_ += length;
		return length;
	}
	unsigned long seek(long offset, SeekOrigin origin)
	{
		switch (origin) {
		case SET:
			break;
		case CUR:
			offset += long(cursor_);
			break;
		case END:
			offset += long(length_);
			break;
		default:
			assert(0);
		}
		if (offset < 0 || long(length_) <= offset) {
			if (offset < 0 || reference_)
				throw MemoryStreamException("seek");
			else
				expand_(offset);
		}
		return static_cast<unsigned long>(cursor_ = offset);
	}
	void clear()
	{
		if (!reference_)
			free(ptr_);
		reference_ = false;
		ptr_       = 0;
		cursor_    = 0;
		length_    = 0;
		capacity_  = 0;
		writeable_ = true;
	}
	virtual ~MemoryStream()
	{
		clear();
	}
protected:
	bool   reference_;
	byte * ptr_;
	size_t cursor_;
	size_t length_;
	size_t capacity_;
	enum { CAPACITY_INIT = 1024 };
	void expand_(size_t n)
	{
		assert(!reference_);
		size_t cap = capacity_ == 0 ? CAPACITY_INIT : capacity_;
		while (cap < n)
			cap *= 2;
		if (cap != capacity_) {
			byte *p = reinterpret_cast<byte *>(realloc(ptr_, cap));
			if (!p)
				throw MemoryStreamException("out of memory");
			ptr_      = p;
			capacity_ = cap;
		}
		length_ = n;
	}
};

#endif	//	_STREAM_INC_
